/*
	ph_.c ver 1.4 (c) PaulHoule.com  10/6/2019

	Implementation of ph_.lib -- see ph_.h for documentation.

	The approach is to create a subset of the C runtime library using
	thin wrappers around Windows OS functions.

	This is optimized for size (not speed) using the MSVC 8 compiler.
	Target use would be a simple command line tool.
*/

#include "ph_.h"

//$$$ ----- would be nice to be able to include <windows.h> again if all the
//$$$ ----- stdlib support <windows.h> needs can be put in "ph.h"
//$$$ #define WIN32_LEAN_AND_MEAN
//$$$ #include <windows.h>
typedef void *HANDLE;
typedef unsigned __int32 DWORD;
typedef long LONG;
typedef int BOOL;
#define FILE_SHARE_READ			0x1
#define FILE_SHARE_WRITE		0x2
#define FILE_ATTRIBUTE_NORMAL	0x80
#define FILE_BEGIN				0
#define FILE_CURRENT			1
#define FILE_END				2
#define ERROR_BROKEN_PIPE		109
#define INVALID_HANDLE_VALUE	((HANDLE) (intptr_t) -1)

#define impwd(rtype) __declspec(dllimport) rtype __stdcall
impwd(void) ExitProcess(unsigned);
impwd(int) wvsprintfA(char *pDst, char *pFmt, va_list vl);
impwd(int) wvsprintfW(wchar_t *pDst, wchar_t *pFmt, va_list vl);
impwd(HANDLE) GetProcessHeap();
impwd(void *) HeapAlloc(HANDLE h, DWORD flags, size_t n);
impwd(BOOL) HeapFree(HANDLE h, DWORD flags, void *pMem);
impwd(HANDLE) GetStdHandle(DWORD nStdHandle);
impwd(HANDLE) CreateFileA(char *pFname, DWORD acc, DWORD shr,
	void *pSecAtt, DWORD crdisp, DWORD flatt, HANDLE hTmpl);
impwd(HANDLE) CreateFileW(wchar_t *pFname, DWORD acc, DWORD shr,
	void *pSecAtt, DWORD crdisp, DWORD flatt, HANDLE hTmpl);
impwd(BOOL) CloseHandle(HANDLE h);
impwd(BOOL) WriteFile(HANDLE h, void *pBuf, DWORD n, DWORD *pBio, void *pOv);
impwd(DWORD) GetLastError();
impwd(BOOL) ReadFile(HANDLE h, void *pBuf, DWORD n, DWORD *pBio, void *pOv);
impwd(DWORD) SetFilePointer(HANDLE h, LONG Dist, LONG *pDistH, DWORD mthd);

#pragma comment(lib, "user32")
#pragma comment(lib, "kernel32")

// XF... can't change -- their values are assumed in some formulas.
#define XFBIN 8
#define XFAPP 4
#define XFERR 2
#define XFEOF 1
typedef struct {
	int32_t hw;
	uint16_t err;
	uint8_t f;
	uint8_t u;
} xfile;
static xfile xfa[16];					// [0] unused, [1..3]= std err,out,in

void __fastcall exit(unsigned ecode) {
	ExitProcess(ecode);
}

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; prnf: In overflow cases some garbage can be output -- the wvsprintfA/W
	; result might not accurately reflect the length of the result string.
	; I'm guessing this only happens when 1024 is returned; have to do some
	; testing to verify that.
	;
	; Note wvsprintf#() can output 1025 (w)chars (1024 and '\0' terminator).
	*/
#define prnf(FN, FMT_T, FUNC_AW)\
int __fastcall FN(FMT_T *pFmt, ...) {\
	FMT_T pDst[1040];\
	va_list vl;\
	int i;\
	va_start(vl, pFmt);\
	i = FUNC_AW(pDst, pFmt, vl);\
	va_end(vl);\
	return (int) fwrite(pDst, sizeof(*pFmt), i, stdout);\
}
prnf(printf, char, wvsprintfA)
prnf(wprintf, wchar_t, wvsprintfW)

#define vprnf(FN, FMT_T, FUNC_AW)\
int __fastcall FN(FMT_T *pFmt, va_list vl) {\
	FMT_T pDst[1040];\
	int i = FUNC_AW(pDst, pFmt, vl);\
	return (int) fwrite(pDst, sizeof(*pFmt), i, stdout);\
}
vprnf(vprintf, char, wvsprintfA)
vprnf(vwprintf, wchar_t, wvsprintfW)

#define fprnf(FN, FMT_T, FUNC_AW)\
int __fastcall FN(FILE *stream, FMT_T *pFmt, ...) {\
	FMT_T pDst[1040];\
	va_list vl;\
	int i;\
	va_start(vl, pFmt);\
	i = FUNC_AW(pDst, pFmt, vl);\
	va_end(vl);\
	return (int) fwrite(pDst, sizeof(*pFmt), i, stream);\
}
fprnf(fprintf, char, wvsprintfA)
fprnf(fwprintf, wchar_t, wvsprintfW)

#define snprnf(FN, FMT_T, FUNC_AW, FUNC_MC)\
int __fastcall FN(FMT_T *pDst, size_t n, FMT_T *pFmt, ...) {\
	FMT_T Tmp[1040];\
	va_list vl;\
	int i;\
	va_start(vl, pFmt);\
	i = FUNC_AW(Tmp, pFmt, vl);\
	if ((unsigned) i < n) n = (unsigned) i + 1;\
	if (n) { pDst[--n] = '\0'; FUNC_MC(pDst, Tmp, n); }\
	va_end(vl);\
	return i;\
}
snprnf(snprintf, char, wvsprintfA, memcpy)
snprnf(swprintf, wchar_t, wvsprintfW, wmemcpy)

#define vsnprnf(FN, FMT_T, FUNC_AW, FUNC_MC)\
int __fastcall FN(FMT_T *pDst, size_t n, FMT_T *pFmt, va_list vl) {\
	FMT_T Tmp[1040];\
	int i = FUNC_AW(Tmp, pFmt, vl);\
	if ((unsigned) i < n) n = (unsigned) i + 1;\
	if (n) { pDst[--n] = '\0'; FUNC_MC(pDst, Tmp, n); }\
	return i;\
}
vsnprnf(vsnprintf, char, wvsprintfA, memcpy)
vsnprnf(vswprintf, wchar_t, wvsprintfW, wmemcpy)

#define vfprnf(FN, FMT_T, FUNC_AW)\
int __fastcall FN(FILE *stream, FMT_T *pFmt, va_list vl) {\
	FMT_T pDst[1040];\
	int i;\
	i = FUNC_AW(pDst, pFmt, (va_list) vl);\
	return (int) fwrite(pDst, sizeof(*pFmt), i, stream);\
}
vfprnf(vfprintf, char, wvsprintfA)
vfprnf(vfwprintf, wchar_t, wvsprintfW)

int __fastcall sprintf(char *pDst, char *pFmt, ...) {
	va_list vl;
	int i;
	va_start(vl, pFmt);
	i = wvsprintfA(pDst, pFmt, vl);
	va_end(vl);
	return i;
}

int __fastcall vsprintf(char *pDst, char *pFmt, va_list vl) {
	return wvsprintfA(pDst, pFmt, (va_list) vl);
}

int __fastcall fputc(int c, FILE *stream) {
	int i = fwrite((char *) &c, 1, 1, stream) < 1 ? EOF : 0;
	return i | c;
}

int __fastcall putchar(int c) {
	return fputc(c, stdout);
}

int __fastcall fputs(char *p, FILE *stream) {
	size_t i = strlen(p);
	return fwrite(p, 1, i, stream) < i ? EOF : 0;
}

int __fastcall puts(char *p) {
	int i = fputs(p, stdout);
	if (!i) i = fputs("\n", stdout);
	return i;
}

void *__fastcall malloc(size_t n) {
	return HeapAlloc(GetProcessHeap(), 0, n);
}

void __fastcall free(void *p) {
	if (p) HeapFree(GetProcessHeap(), 0, p);
}

#define isxx(FN, C_T, TMP_T, OR, SC, CNT)\
int __fastcall FN(C_T c) {\
	return (TMP_T) (((TMP_T) c | OR) - SC) < CNT ? -1 : 0;\
}
isxx(isdigit, int, uint8_t, 0, '0', 10)
isxx(iswdigit, wint_t, __ph_uwchar_t, 0, '0', 10)
isxx(isalpha, int, uint8_t, 0x20, 'a', 26)
isxx(iswalpha, wint_t, __ph_uwchar_t, 0x20, 'a', 26)
isxx(islower, int, uint8_t, 0, 'a', 26)
isxx(iswlower, wint_t, __ph_uwchar_t, 0, 'a', 26)
isxx(isupper, int, uint8_t, 0, 'A', 26)
isxx(iswupper, wint_t, __ph_uwchar_t, 0, 'A', 26)

#ifndef PH_ASM	// ----- if not using ph_.asm, use C equivalents -----

#define issp(FN, C_T, TMP_T)\
int __fastcall FN(C_T c) {\
	uint32_t cc = (TMP_T) c;\
	if (cc != ' ') cc = (cc - 9) < 5 ? -1 : 0;\
	return cc;\
}
issp(isspace, int, uint8_t)
issp(iswspace, wint_t, __ph_uwchar_t)

#define mcmp(FN, SRC_DST_T, UNS_T)\
int __fastcall FN(SRC_DST_T *pDst, SRC_DST_T *pSrc, size_t n) {\
	size_t nc = n + 1;\
	intptr_t s_d = (char *) pSrc - (char *) pDst;\
	int dc = 0, sc;\
	do {\
		if (!--nc) break;\
		dc = *(UNS_T *) pDst, sc = *(UNS_T *) ((char *) pDst + s_d), pDst++;\
	} while (!(dc -= sc));\
	return dc;\
}
mcmp(memcmp, char, uint8_t)
mcmp(wmemcmp, wchar_t, wchar_t)

#if 0 // ?memcpy() unused -- ?memcpy #define'd to ?memmove in ph_.h
#define mcpy(FN, SRC_DST_T, TMP_T)\
SRC_DST_T *__fastcall FN(SRC_DST_T *pDst, SRC_DST_T *pSrc, size_t n) {\
	size_t nc = n;\
	TMP_T *pD = (TMP_T *) pDst, *pS = (TMP_T *) pSrc;\
	if (nc) do { *pD++ = *pS++; } while (--nc);\
	return pDst;\
}
mcpy(memcpy, void, char)
mcpy(wmemcpy, wchar_t, wchar_t)
#endif

#define mmov(FN, SRC_DST_T, TMP_T)\
SRC_DST_T *__fastcall FN(SRC_DST_T *pDst, SRC_DST_T *pSrc, size_t n) {\
	size_t nc = n;\
	size_t dir = (size_t) 1;\
	TMP_T *pD = (TMP_T *) pDst, *pS = (TMP_T *) pSrc;\
	if (pD > pS) pD += nc - 1, pS += nc - 1, dir = -(intptr_t) dir;\
	if (nc) do { *pD = *pS, pD += dir, pS += dir; } while (--nc);\
	return pDst;\
}
mmov(memmove, void, char)
mmov(wmemmove, wchar_t, wchar_t)

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; Had to be careful to write memset in such a way that MSVC compiler
	; didn't replace the loop with a call to memset (which doesn't exist
	; in a build without the MSVC CRT lib).
	*/
#define mset(FN, DST_T, C_T, TMP_T)\
DST_T *__fastcall FN(DST_T *pDst, C_T c, size_t n) {\
	size_t nc = n;\
	if (nc) {\
		TMP_T *pD = pDst;\
		do { *pD++ = (TMP_T) c; } while (--nc);\
	}\
	return pDst;\
}
mset(memset, void, int, char)
mset(wmemset, wchar_t, wchar_t, wchar_t)

#define mchr(FN, DST_T, C_T, TMP_T)\
DST_T *__fastcall FN(DST_T *pDst, C_T c, size_t n) {\
	size_t nc = n;\
	if (nc) {\
		TMP_T *pD = pDst;\
		do { if (*pD == (TMP_T) c) return pD; pD++; } while (--nc);\
	}\
	return NULL;\
}
mchr(memchr, void, int, char)
mchr(wmemchr, wchar_t, wchar_t, wchar_t)

#define slen(FN, SRC_T, TMP_T)\
size_t __fastcall FN(SRC_T *pSrc) {\
	SRC_T *pS = pSrc;\
	do {} while (*pS++);\
	return (pS - pSrc) - 1;\
}
slen(strlen, char, char)
slen(wcslen, wchar_t, uint32_t)

#define scpy(FN, SRC_DST_T)\
SRC_DST_T *__fastcall FN(SRC_DST_T *pDst, SRC_DST_T *pSrc) {\
	SRC_DST_T *pD = pDst;\
	do {} while (*pD++ = *pSrc++);\
	return pDst;\
}
scpy(strcpy, char)
scpy(wcscpy, wchar_t)

#define sncpy(FN, SRC_DST_T, TMP_T)\
SRC_DST_T *__fastcall FN(SRC_DST_T *pDst, SRC_DST_T *pSrc, size_t n) {\
	SRC_DST_T *pD = pDst;\
	size_t nc = n;\
	if (nc) {\
	  TMP_T c = -1;\
	  do { if (c) c = *pSrc++; *pD++ = (SRC_DST_T) c; } while (--nc);\
	}\
	return pDst;\
}
sncpy(strncpy, char, char)
sncpy(wcsncpy, wchar_t, uint32_t)

#define scmp(FN, SRC_DST_T, UNS_T)\
int __fastcall FN(SRC_DST_T *pDst, SRC_DST_T *pSrc) {\
	int dc, sc;\
	do { dc = *(UNS_T *) pDst++, sc = *(UNS_T *) pSrc++; }\
	  while (!(dc -= sc) && sc);\
	return dc;\
}
scmp(strcmp, char, uint8_t)
scmp(wcscmp, wchar_t, wchar_t)

#define scat(FN, SRC_DST_T, TMP_T)\
SRC_DST_T *__fastcall FN(SRC_DST_T *pDst, SRC_DST_T *pSrc) {\
	SRC_DST_T *pD = pDst;\
	TMP_T c;\
	if (*pSrc) {\
	  do { c = *pD++; } while (c);\
	  pD--;\
	  do { c = *pSrc++; *pD++ = (SRC_DST_T) c; } while (c);\
	}\
	return pDst;\
}
scat(strcat, char, char)
scat(wcscat, wchar_t, uint32_t)

#define sncat(FN, SRC_DST_T, TMP_T)\
SRC_DST_T *__fastcall FN(SRC_DST_T *pDst, SRC_DST_T *pSrc, size_t n) {\
	SRC_DST_T *pD = pDst;\
	size_t nc = n;\
	if (*pSrc && nc) {\
	  TMP_T c;\
	  do { c = *pD++; } while (c);\
	  pD--;\
	  do { c = *pSrc++; if (!c) break; *pD++ = (SRC_DST_T) c; } while (--nc);\
	  *pD++ = '\0';\
	}\
	return pDst;\
}
sncat(strncat, char, char)
sncat(wcsncat, wchar_t, uint32_t)

#define schr(FN, SRC_T, C_T, TMP_T)\
SRC_T *__fastcall FN(SRC_T *pSrc, C_T c) {\
	TMP_T sc;\
	do {\
		sc = *pSrc++;\
		if (sc == (TMP_T) c) return pSrc - 1;\
	} while (sc);\
	return NULL;\
}
schr(strchr, char, int, char)
schr(wcschr, wchar_t, wchar_t, uint32_t)

#endif						// ----- end of #ifndef PH_ASM -----

#define srchr(FN, SRC_T, C_T, TMP_T)\
SRC_T *__fastcall FN(SRC_T *pSrc, C_T c) {\
	TMP_T sc;\
	SRC_T *pC = NULL;\
	do {\
		sc = *pSrc++;\
		if (sc == (TMP_T) c) pC = pSrc - 1;\
	} while (sc);\
	return pC;\
}
srchr(strrchr, char, int, char)
srchr(wcsrchr, wchar_t, wchar_t, uint32_t)

#define sncmp(FN, SRC_DST_T, UNS_T)\
int __fastcall FN(SRC_DST_T *pDst, SRC_DST_T *pSrc, size_t n) {\
	size_t nc = n + 1;\
	intptr_t s_d = (char *) pSrc - (char *) pDst;\
	int dc = 0, sc;\
	do {\
		if (!--nc) break;\
		dc = *(UNS_T *) pDst, sc = *(UNS_T *) ((char *) pDst + s_d), pDst++;\
	} while (!(dc -= sc) && sc);\
	return dc;\
}
sncmp(strncmp, char, uint8_t)
sncmp(wcsncmp, wchar_t, wchar_t)

#define sstr(FN, SRC_DST_T, TMP_T)\
SRC_DST_T *__fastcall FN(SRC_DST_T *pDst, SRC_DST_T *pSrc) {\
	TMP_T dc, sc;\
	do {\
	  SRC_DST_T *pD = pDst++, *pS = pSrc;\
	  do { dc = *pD++, sc = *pS++; } while (sc && sc == dc);\
	  if (sc && !dc) return NULL;\
	} while (sc);\
	return pDst - 1;\
}
sstr(strstr, char, char)
sstr(wcsstr, wchar_t, uint32_t)

/*----------------------------------------------------------------------------
;
; Translates incoming 'stream' (which is a xfa index) to ptr to xfa
; entry.  Also, if not already done, inits entry ->hw for std streams.
;
;---------------------------------------------------------------------------*/

static xfile *__fastcall xfptr(FILE *stream) {
	xfile *pX = &xfa[(intptr_t) stream];
	if ((size_t) stream <= 3			// if need to init in/out/err (1/2/3),
	  && stream && !pX->hw) {
			// assumes		: #define STD_INPUT_HANDLE  (DWORD)-10
			// windows		: #define STD_OUTPUT_HANDLE (DWORD)-11
			// definitions	: #define STD_ERROR_HANDLE  (DWORD)-12
		pX->hw = (int32_t) (intptr_t)
		  GetStdHandle((uint32_t) -((intptr_t) stream + 9));
	}
	return pX;
}

void __fastcall clearerr(FILE *stream) { // clear handle error/eof flags
	xfa[(intptr_t) stream].f &= ~(XFERR | XFEOF);
}

int __fastcall ferror(FILE *stream) {	// nz if file has error status
	return xfa[(intptr_t) stream].f & XFERR;
}

int __fastcall feof(FILE *stream) {		// nz if file has EOF status
	return xfa[(intptr_t) stream].f & XFEOF;
}

/*----------------------------------------------------------------------------
;
; fopen and _wfopen (wide char _wfopen is Microsoft-specific).
;
;---------------------------------------------------------------------------*/

#if 0 // 10/4/2019: can't check values -- no longer including windows.h
// sanity check fopen constants; also helps document their use in code
#if GENERIC_READ != 1<<31 || GENERIC_WRITE != 1<<30 \
	|| CREATE_ALWAYS != 2 || OPEN_EXISTING != 3 || OPEN_ALWAYS != 4
#error unexpected windows.h CreateFile-related constant(s)
#endif
#endif

#define fopen_fhdr(FN, C_T)	/* generate proto, declare locals */\
FILE * __fastcall FN(C_T *fn, C_T *m) {\
	HANDLE hw;\
	unsigned iX;\
	uint8_t f;

#define fopen_set_f(TMP_T)	/* parses m to set f (flags) */\
	do {\
		TMP_T c = *m;\
		f = (2 << 5) | 1;				/* assume/check-for 'w' */\
		if (c == 'w') break;\
		f = (3 << 5) | 2;				/* assume/check-for 'r' */\
		if (c == 'r') break;\
		f = (4 << 5) | XFAPP | 1;		/* assume/check-for 'a' */\
		if (c == 'a') break;\
		return NULL;\
	} while (0);\
	while (1) {\
		TMP_T c = *++m;\
		if (!c) break;\
		if (c == 'b') { f |= XFBIN; continue; }\
		if (c == '+') { f |= 3; continue; }\
		return NULL;\
	}

#define fopen_set_iX			/* sets iX to 1st free file table index */\
	iX = 3;\
	do {\
		if (++iX == sizeof(xfa) / sizeof(*xfa)) return NULL;\
	} while (xfa[iX].hw);

#define fopen_set_hw(A_or_W)	/* calls CreateFile? to set hw */\
	hw = CreateFile ## A_or_W(\
		fn,									/* file name */\
		(DWORD) f << 30,					/* access mode */\
		FILE_SHARE_READ | FILE_SHARE_WRITE, /* share mode (allow r/w) */\
		NULL,								/* default security attributes */\
		f >> 5,								/* create disposition (2,3,4) */\
		FILE_ATTRIBUTE_NORMAL,				/* flags and attributes */\
		NULL								/* template file */\
	);

#define fopen_ftrl			/* init *pX entry and return */\
	if (hw == INVALID_HANDLE_VALUE) return NULL;\
	xfa[iX].hw = (int32_t) (intptr_t) hw;\
	xfa[iX].f = f & (XFAPP | XFBIN);\
	return (FILE *) (intptr_t) iX;\
}

fopen_fhdr(fopen, char)
	fopen_set_f(char)
	fopen_set_iX
	fopen_set_hw(A)
	fopen_ftrl

fopen_fhdr(_wfopen, wchar_t)
	fopen_set_f(uint32_t)
	fopen_set_iX
	fopen_set_hw(W)
	fopen_ftrl

// sanity check mode constants
#if _O_TEXT != 0 || _O_BINARY != 8 || XFBIN != _O_BINARY
#error unexpected _O_BINARY/_O_TEXT/XFBIN constant(s)
#endif

int __fastcall _setmode(int fd, int mode) {
	uint8_t *pF = &xfa[fd].f;			// ptr to file table flags
	int rc = *pF;						// rc= current flags (all flags)

	mode = (mode ^ rc) & XFBIN;			// mode= XFBIN set if mode change
	*pF = (uint8_t) (rc ^ mode);		// store new flags (perhaps no change)
	return rc & XFBIN;					// return previous bin/text setting
}

int __fastcall fclose(FILE *stream) {
	HANDLE hw;
	xfile *pX = xfptr(stream);
	hw = (HANDLE) (size_t) pX->hw;
	*(uint64_t *) pX &= 0;				// zero (re-init) file array entry
	return CloseHandle(hw) ? 0 : EOF;
}

/*----------------------------------------------------------------------------
;
; Filter to handle "\n"->"\r\n" expansion when writing in text mode.
; Always sets *pBio = n, which is invalid if error (zero) return status.
;
;---------------------------------------------------------------------------*/

typedef BOOL (__stdcall *WIN_IO_FN)(HANDLE, void *, DWORD, DWORD *, void *);

static BOOL __stdcall WriteText(HANDLE hw, uint8_t *p,
	DWORD n, DWORD *pBio, void *pOv
) {
	uint8_t *pEnd = p + n;			// end of source data (n may be 0)
	DWORD bio;
	BOOL rc;
	uint8_t buf[120];				// block of translated output data

	*pBio = n;
	do {
		uint8_t *pBuf = buf;			// prepare to write translated bytes
		do {
			uint8_t c;
			if (p == pEnd) break;		// break= no more source bytes
			c = *p++;					// translate next source byte
			if (c == '\n') *pBuf++ = '\r';
			*pBuf++ = c;
		} while (pBuf <= &buf[sizeof(buf) - 2]); // loop if more buf room
		rc = WriteFile(hw, buf, (DWORD) (pBuf - buf), &bio, NULL);
	} while (rc && p != pEnd);			// loop in no error and more source
	return rc;
}

/*----------------------------------------------------------------------------
;
; Byte-level i/o (below element-level).  At this point the i/o size has been
; resolved to a DWORD count of bytes, that must NOT be 0 or (DWORD) -1.
;
; Returns count of bytes i/o'd, or (DWORD) -1 for error.
; If returned value < requested bytes, EOF has been hit.
;
; The XFERR or XFEOF bit is set if either even occurs.
;
;---------------------------------------------------------------------------*/

static DWORD __fastcall io_bytes(
	uint8_t *p, DWORD bytes, xfile *pX, WIN_IO_FN pFn
) {
	DWORD totbio = 0;					// total bytes completed counter
	do {
		DWORD bio;						// bytes i/o'd (return value)
		if (!pFn(						// execute i/o -- if error,
		  (HANDLE) (size_t) pX->hw, p, bytes, &bio, NULL)) {
			DWORD e = GetLastError();	// get Windows error code
			if (e != ERROR_BROKEN_PIPE) { // if not broken pipe (that's EOF),
				pX->err = (uint16_t) e; // save error code
				pX->f |= XFERR;			// set error stream status
				totbio = (DWORD) -1;	// set error return status
				break;
			}
		}
		if (!bio) {						// if 0 bytes returned,
			pX->f |= XFEOF;				// set EOF status
			break;
		}
		totbio += bio;					// update total, ptr, left-to-do
		p += bio;
		bytes -= bio;
	} while (bytes);					// i/o again if incomplete
	return totbio;
}

/*----------------------------------------------------------------------------
;
; Element-level i/o (above byte-level).  At this point we're dealing with an
; i/o of n elements each of size sz bytes.  The i/o could be for 0 bytes.
;
; Note limitation: sz must be < (DWORD) -1 (ie. 0xfffffffe or less).
;
;---------------------------------------------------------------------------*/

static size_t __fastcall io_elems(
	uint8_t *p, size_t sz, size_t n, xfile *pX, WIN_IO_FN pFn
) {
	size_t teio = 0;					// total elements successfully i/o'd
	size_t mep = 0;					// max elements per i/o

	if (sz) {							// if (perhaps) something to i/o,
		if (sz == (DWORD) sz			// can't handle sz >= (DWORD) -1
		 && (DWORD) sz < (DWORD) -1) {
			mep = (16*1024*1024) / sz;	// calc max elements per i/o
			if (!mep) mep++;			// (must be at least 1)
		}
	}

	do {
		DWORD r;
		if (mep > n) mep = n;			// limit to total elements left
		if (!mep) break;
		r = io_bytes(p, (DWORD) (mep * sz), pX, pFn);
		if (r == (DWORD) -1) break;		// break= i/o error
		p += r;							// advance p in case we're not done
		r /= (DWORD) sz;				// r = count of elements transferred
		teio += r;						// update total elements transferred
		if (r != mep) break;			// break= not full xfer
	} while (n -= mep);

	return teio;
}

size_t __fastcall fread(void *p, size_t sz, size_t n, FILE *stream) {
	xfile *pX = xfptr(stream);

	if (!(pX->f & XFBIN)) {				// set EOF if (unsupported) text mode
		pX->f |= XFEOF;
	}
	if (pX->f & XFEOF) return 0;		// do nothing EOF status exists
	return io_elems(p, sz, n, pX, ReadFile);
}

size_t __fastcall fwrite(void *p, size_t sz, size_t n, FILE *stream) {
	xfile *pX = xfptr(stream);

	if (pX->f & XFAPP) {				// if opened as 'a' (append),
		SetFilePointer(					// position to EOF before write
			(HANDLE) (size_t) pX->hw, 0, NULL, FILE_END);
	}
	return io_elems(p, sz, n, pX, pX->f & XFBIN
		? (WIN_IO_FN) WriteFile : (WIN_IO_FN) WriteText);
}

// sanity check offset constants
#if FILE_BEGIN != SEEK_SET || FILE_CURRENT != SEEK_CUR \
	|| FILE_END != SEEK_END
#error unexpected windows.h SetFilePointer-related constant(s)
#endif

int __fastcall fseek(FILE *stream, long off, int orig) {
	xfile *pX = xfptr(stream);

	pX->f &= ~XFEOF;					// seeking clears EOF state
	SetFilePointer((HANDLE) (size_t) pX->hw, off, NULL, orig);
	return 0;
}

long __fastcall ftell(FILE *stream) {
	xfile *pX = xfptr(stream);
	return SetFilePointer((HANDLE) (size_t) pX->hw, 0, NULL, SEEK_CUR);
}
